use std::{env, fs};
use std::fs::{File, Permissions};
use std::io::Write;
use std::os::unix::fs::PermissionsExt;
use std::path::Path;
use kube::{Api, Client};
use log::{info, debug};
use rand::rngs::OsRng;
use ssh_key::{PrivateKey, Algorithm, LineEnding};
use k8s_openapi::api::core::v1::{ConfigMap, Node, Pod};
use kube::api::{Patch, PatchParams};
use serde_json::json;
use crate::api::cluster::volume::create_volume_unmanaged;
use crate::api::entity::nodes::P5xError;
use crate::api::services::Services;

pub async fn ensure_system_disk(svc: &Services<'_>) -> Result<(), P5xError> {
    info!(target: "p5x", "Ensuring that the P5x API system disk exists...");

    // Load the dynamic-kv
    let client = Client::try_default().await.map_err(P5xError::KubeError)?;
    let namespace = fs::read_to_string("/var/run/secrets/kubernetes.io/serviceaccount/namespace")
        .unwrap_or_else(|_| "p5x-system".to_string());

    let maps: Api<ConfigMap> = Api::namespaced(client.clone(), &namespace);
    let map = maps.get("dynamic-kv").await.map_err(P5xError::KubeError)?;

    // If there's already a PVE host and disk mount in the config, we're done
    let mut data = map.data.unwrap_or_default();
    if let Some(host) = data.get("api-pve-host") {
        if let Some(mount) = data.get("api-pve-disk") {
            info!(target: "p5x", "Found existing P5x API system disk on {host}:{mount}");
            return Ok(());
        }
    }

    // Otherwise, create the disk on this PVE host
    info!(target: "p5x", "Provisioning new P5x API system disk (this is a one-time fixup)...");

    // Load the labels for this pod's node
    let pod_name = env::var("POD_NAME").expect("Could not determine POD_NAME from environment!");
    let pods: Api<Pod> = Api::namespaced(client.clone(), &namespace);
    let pod = pods.get(&pod_name).await.map_err(P5xError::KubeError)?;

    let node_name = pod.spec.and_then(|spec| spec.node_name).expect("Could not determine the Node name for pod!");

    let nodes: Api<Node> = Api::all(client);
    let node = nodes.get(&node_name).await.map_err(P5xError::KubeError)?;
    let labels = node.metadata.labels.expect("Could not load labels for node");

    let pve_host = labels.get("p5x.garrettmills.dev/pve-host")
        .expect("Node is missing required label: p5x.garrettmills.dev/pve-host");

    let pve_id: i32 = labels.get("p5x.garrettmills.dev/pve-id")
        .expect("Node is missing required label: p5x.garrettmills.dev/pve-id")
        .parse()
        .unwrap();

    debug!(target: "p5x", "pod {pod_name} | node {node_name} | host {pve_host} | vmid {pve_id}");

    // Create a new disk and mount it to the K8s node (LXC container)
    let (disk_name, _) = create_volume_unmanaged(
        svc,
        5 * 1024 * 1024 * 1024,
        pve_host,
        pve_id,
        "p5x-api-system-disk"
    ).await?;

    // Add it to the dynamic-kv config and save
    data.insert("api-pve-host".to_string(), pve_host.to_string());
    data.insert("api-pve-disk".to_string(), disk_name);
    let patch = json!({
        "data": data
    });

    let params = PatchParams::apply("p5x-api");
    maps.patch("dynamic-kv", &params, &Patch::Merge(&patch))
        .await.map_err(P5xError::KubeError)?;

    info!(target: "p5x", "Successfully provisioned P5x API system disk.");
    Ok(())
}


/** Check if the SSH pubkey/privkey exist at the configured paths. If not, generate them. */
pub fn ensure_ssh_keypair() -> Result<(), ssh_key::Error> {
    let pubkey_path = env::var("P5X_SSH_PUBKEY_PATH").expect("Missing env: P5X_SSH_PUBKEY_PATH");
    let privkey_path = env::var("P5X_SSH_PRIVKEY_PATH").expect("Missing env: P5X_SSH_PRIVKEY_PATH");

    // If both exist, then p5x will boot correctly when it reads from the files.
    // If only one of the two files exists, p5x will error on boot. This is safer
    // than accidentally overwriting someone's key.
    if Path::new(&pubkey_path).exists() || Path::new(&privkey_path).exists() {
        info!(target: "p5x", "Found existing SSH keypair on filesystem");
        return Ok(());
    }

    // Generate an ed25519 keypair
    info!(target: "p5x", "Generating a new SSH keypair");
    let mut csprng = OsRng;
    let privkey = PrivateKey::random(&mut csprng, Algorithm::Ed25519)?;

    // Write the privkey to a file
    let privkey_pem = privkey.to_openssh(LineEnding::LF)?;
    let mut privkey_file = File::create(privkey_path)?;
    privkey_file.write_all(privkey_pem.as_bytes())?;
    privkey_file.set_permissions(Permissions::from_mode(0o600))?;

    // Write the pubkey to a file
    let pubkey = privkey.public_key();
    let pubkey_ssh = pubkey.to_openssh()?;
    let mut pubkey_file = File::create(pubkey_path)?;
    pubkey_file.write_all(pubkey_ssh.as_bytes())?;

    Ok(())
}
